from django.contrib.auth.models import Permission
from django.contrib.auth.models import User
from django.test import TestCase

from django_delayed_union import DelayedUnionQuerySet

from .factories import UserFactory
from .mixins import DelayedQuerySetMetaTestsMixin
from .mixins import DelayedQuerySetTestsMixin


class DelayedUnionQuerySetMetaTests(DelayedQuerySetMetaTestsMixin, TestCase):
    def get_class(self):
        return DelayedUnionQuerySet

    def test_delayed_union_only_accepts_all_as_kwarg(self):
        with self.assertRaises(TypeError):
            DelayedUnionQuerySet(
                User.objects.all(),
                User.objects.all(),
                foo=42
            )


class DelayedUnionQuerySetTestsMixin(DelayedQuerySetTestsMixin):
    def test_select_related(self):
        base_qs = Permission.objects.all()
        qs = DelayedUnionQuerySet(base_qs, base_qs)

        self.assertTrue(qs.exists())
        for permission in qs.select_related('content_type'):
            with self.assertNumQueries(0):
                self.assertIsNotNone(permission.content_type.id)

    def test_update(self):
        self.qs.update(first_name='Rover')
        for user in self.qs:
            self.assertEqual(user.first_name, 'Rover')


class DelayedUnionQuerySetTests(DelayedUnionQuerySetTestsMixin, TestCase):
    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.all(),
            User.objects.all(),
        )


class DelayedUnionQuerySetMixedTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.filter(id=self.bad_id),
            User.objects.all(),
        )


class DelayedUnionQuerySetReversedMixedTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.all(),
            User.objects.filter(id=self.bad_id),
        )


class DelayedUnionAllMutuallyExclusiveQuerySetTests(
        DelayedUnionQuerySetTestsMixin,
        TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.filter(id=self.user.id),
            User.objects.exclude(id=self.user.id),
            all=True
        )


class DelayedUnionAllQuerySetTests(DelayedUnionQuerySetTestsMixin, TestCase):

    def get_queryset(self):
        return DelayedUnionQuerySet(
            User.objects.all(),
            User.objects.all(),
            all=True
        )

    def test_get(self):
        with self.assertRaises(User.MultipleObjectsReturned):
            self.qs.get(id=self.user.id)

    def test_repr(self):
        self.assertEqual(
            repr(self.qs),
            '<QuerySet {}>'.format([self.user, self.user])
        )

    def test_count(self):
        self.assertEqual(self.qs.count(), 2)

    def test_iter(self):
        self.assertEqual(list(self.qs), [self.user, self.user])

    def test_order_by(self):
        second_user = UserFactory.create()
        self.assertEqual(
            list(self.qs.order_by('pk')),
            [self.user] * 2 + [second_user] * 2
        )

    def test_order_by_reversed(self):
        second_user = UserFactory.create()
        self.assertEqual(
            list(self.qs.order_by('-pk')),
            [second_user] * 2 + [self.user] * 2
        )

    def test_reverse(self):
        second_user = UserFactory.create()
        self.assertEqual(
            list(self.qs.order_by('pk').reverse()),
            [second_user] * 2 + [self.user] * 2
        )

    def test_values(self):
        self.assertEqual(
            list(self.qs.values('id')),
            [{'id': self.user.id}, {'id': self.user.id}]
        )

    def test_values_list(self):
        self.assertEqual(
            list(self.qs.values_list('id', flat=True)),
            [self.user.id, self.user.id]
        )

    def test_len(self):
        self.assertEqual(len(self.qs), 2)

    def test_iterator(self):
        self.assertEqual(list(self.qs.iterator()), [self.user] * 2)



from .base import DelayedQuerySet


class DelayedIntersectionQuerySet(DelayedQuerySet):
    def __init__(self, *querysets):
        return super(DelayedIntersectionQuerySet, self).__init__(*querysets)

    def _apply_operation(self):
        """
        Returs the union of all of the component querysets.
        """
        return self._querysets[0].intersection(*self._querysets[1:])

    def distinct(self):
        """
        Returns a new :class:`DelayedIntersectionQuerySet` instance that will
        select only distinct results.
        """
        clone = self._clone()
        return clone



import abc
from builtins import object

from django.contrib.auth.models import User
from django.db.models import F
from django.db.models import Q
from django.db.models import QuerySet
from future.utils import with_metaclass

from django_delayed_union.base import DelayedQuerySetDescriptor

from .factories import UserFactory


class DelayedQuerySetMetaTestsMixin(with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def get_class(self):
        """
        Returns the subclass of :class:`DelayedQuerySet` to test.
        """

    def test_matches_public_api_of_queryset(self):
        cls = self.get_class()
        qs_attrs = set(dir(QuerySet))
        dqs_attrs = set(dir(cls))
        missing = filter(
            lambda attr: not attr.startswith('_'),
            qs_attrs - dqs_attrs
        )
        self.assertFalse(
            set(),
            (
                'The following attributes are missing'
                ' from {}: {}'
            ).format(cls, ', '.join(sorted(missing)))
        )

    def test_currently_does_not_support_nested_delayed_querysets(self):
        cls = self.get_class()
        with self.assertRaises(ValueError):
            cls(
                User.objects.all(),
                cls(User.objects.all(), User.objects.all())
            )

    def test_all_descriptors_have_docstrings(self):
        cls = self.get_class()
        for attr in dir(cls):
            descriptor = getattr(cls, attr)
            if not isinstance(descriptor, DelayedQuerySetDescriptor):
                continue
            self.assertIsNotNone(descriptor.__doc__)


class DelayedQuerySetTestsMixin(with_metaclass(abc.ABCMeta, object)):
    @classmethod
    def setUpTestData(cls):
        super(DelayedQuerySetTestsMixin, cls).setUpTestData()
        cls.user = UserFactory.create()
        cls.bad_id = cls.user.id - 1

    def setUp(self):
        super(DelayedQuerySetTestsMixin, self).setUp()
        self.regular_qs = User.objects.filter(id=self.user.id)
        self.qs = self.get_queryset()

    @abc.abstractmethod
    def test_select_related(self):
        """
        Subclasses should implement a test that select_related works.
        """

    def test_get(self):
        self.assertEqual(self.qs.get(id=self.user.id), self.user)

    def test_get_does_not_exist(self):
        with self.assertRaises(User.DoesNotExist):
            self.qs.get(id=self.bad_id)

    def test_get_multiple_objects_returned(self):
        UserFactory.create()
        with self.assertRaises(User.MultipleObjectsReturned):
            self.qs.get(id__gt=-1)

    def test_get_filtered_with_bad_id(self):
        with self.assertRaises(User.DoesNotExist):
            self.qs.filter(id=self.bad_id).get(id=self.user.id)

    def test_repr(self):
        self.assertEqual(repr(self.qs), repr(self.regular_qs))

    def test_contains_true(self):
        self.assertIn(self.user, self.qs)

    def test_contains_false(self):
        self.assertNotIn(self.user, self.qs.filter(id=self.bad_id))

    def test_bool_true(self):
        self.assertTrue(self.qs)

    def test_bool_false(self):
        self.assertFalse(self.qs.filter(id=self.bad_id))

    def test_getitem(self):
        self.assertEqual(self.qs[0], self.user)

    def test_count(self):
        self.assertEqual(self.qs.count(), 1)

    def test_iter(self):
        self.assertEqual(list(self.qs), [self.user])

    def test_filter(self):
        self.assertEqual(self.qs.filter(id=self.bad_id).count(), 0)

    def test_exclude(self):
        self.assertEqual(self.qs.exclude(id=self.user.id).count(), 0)

    def test_filtering_after_ordering(self):
        second = UserFactory.create()
        user = self.qs.order_by('-pk').exclude(id=self.bad_id).first()
        self.assertEqual(user, second)

    def test_values(self):
        self.assertEqual(
            list(self.qs.values('id')),
            [{'id': self.user.id}]
        )

    def test_values_list(self):
        self.assertEqual(
            list(self.qs.values_list('id', flat=True)),
            [self.user.id]
        )

    def test_prefetch_related(self):
        for user in self.qs.prefetch_related('groups'):
            with self.assertNumQueries(0):
                self.assertIsNotNone(user.groups.count())

    def test_prefetch_related_preserves_ordering(self):
        second = UserFactory.create()
        qs = self.qs.order_by('-pk').prefetch_related('groups')
        self.assertEqual(qs.first(), second)
        for user in qs:
            with self.assertNumQueries(0):
                self.assertIsNotNone(user.groups.count())

    def test_order_by(self):
        second = UserFactory.create()
        self.assertEqual(
            list(self.qs.order_by('id')),
            [self.user, second]
        )

    def test_order_by_reversed(self):
        second = UserFactory.create()
        self.assertEqual(
            list(self.qs.order_by('-id')),
            [second, self.user]
        )

    def test_ordered_false(self):
        self.assertFalse(self.qs.ordered)

    def test_ordered_true(self):
        self.assertTrue(self.qs.order_by('id').ordered)

    def test_reverse(self):
        second = UserFactory.create()
        self.assertEqual(
            list(self.qs.order_by('-id').reverse()),
            [self.user, second]
        )

    def test_len(self):
        self.assertEqual(len(self.qs), 1)

    def test_nonzero_true(self):
        self.assertTrue(self.qs)

    def test_nonzero_false(self):
        self.assertFalse(self.qs.filter(id=self.bad_id))

    def test_iterator(self):
        self.assertEqual(list(self.qs.iterator()), [self.user])

    def test_earliest(self):
        UserFactory.create()
        self.assertEqual(self.qs.earliest('pk'), self.user)

    def test_latest(self):
        second = UserFactory.create()
        self.assertEqual(self.qs.latest('pk'), second)

    def test_first(self):
        UserFactory.create()
        self.assertEqual(self.qs.first(), self.user)

    def test_first_with_ordering(self):
        second = UserFactory.create()
        self.assertEqual(self.qs.order_by('-pk').first(), second)

    def test_last(self):
        second = UserFactory.create()
        self.assertEqual(self.qs.last(), second)

    def test_last_with_ordering(self):
        UserFactory.create()
        self.assertEqual(self.qs.order_by('-pk').last(), self.user)

    def tests_exists_true(self):
        self.assertTrue(self.qs.exists())

    def tests_exists_false(self):
        self.assertFalse(self.qs.filter(id=self.bad_id).exists())

    def test_none(self):
        self.assertEqual(list(self.qs.none()), [])

    def test_db(self):
        self.assertEqual(self.qs.db, 'default')

    def test_annotate(self):
        user = self.qs.annotate(doubled=2 * F('id')).first()
        self.assertEqual(user.doubled, 2 * self.user.id)

    def test_only(self):
        user = self.qs.only('id').first()
        with self.assertNumQueries(0):
            self.assertEqual(user.id, self.user.id)
        with self.assertNumQueries(1):
            self.assertEqual(user.date_joined, self.user.date_joined)

    def test_defer(self):
        user = self.qs.defer('date_joined').first()
        with self.assertNumQueries(0):
            self.assertEqual(user.id, self.user.id)
        with self.assertNumQueries(1):
            self.assertEqual(user.date_joined, self.user.date_joined)

    def test_in_bulk(self):
        self.assertEqual(self.qs.in_bulk(), {self.user.id: self.user})

    def test_in_bulk_empty(self):
        self.assertEqual(self.qs.in_bulk([]), {})

    def test_in_bulk_with_ids(self):
        self.assertEqual(
            self.qs.in_bulk([self.user.id, self.bad_id]),
            {self.user.id: self.user}
        )

    def test_distinct(self):
        self.assertEqual(list(self.qs.distinct()), [self.user])

    def test_extra(self):
        user = self.qs.extra({'n': 'SELECT 42'}).first()
        self.assertEqual(user.n, 42)

    def test_using(self):
        qs = self.qs.using('some_db')
        self.assertEqual(qs.db, 'some_db')

    def test_create(self):
        user = self.qs.create()
        self.assertIsNotNone(user.id)

    def test_bulk_create(self):
        self.qs.bulk_create([User(id=4242)])
        self.assertTrue(User.objects.filter(id=4242).exists())

    def test_complex_filter(self):
        qs = self.qs.complex_filter(Q(id=-1))
        self.assertEqual(qs.count(), 0)

    def test_get_or_create(self):
        with self.assertRaises(NotImplementedError):
            self.qs.get_or_create(id=4242)



import abc
import inspect
from builtins import object
from functools import partial

from django.db.models import QuerySet
from future.utils import with_metaclass

from .utils import get_formatted_function_signature


class DelayedQuerySetDescriptor(with_metaclass(abc.ABCMeta, object)):
    """
    A base class for the descriptors which are used for
    :class:`DelayedQuerySet`.

    Used in conjunction with :class:`DelayedQuerySetBase` in order to set
    their names upon class creation.
    """
    def __init__(self, name=None):
        self.set_name(name)

    def set_name(self, name):
        """
        Sets the name of the descriptor to *name*.

        :param str name: the name of the descriptor
        """
        self.name = name
        if name is not None:
            self.__doc__ = self.get_docstring()

    @abc.abstractmethod
    def get_base_docstring(self):
        """
        Returns a string which will be prepended to the docstring for
        the corresponding method on :class:`django.db.models.QuerySet`.
        """

    def get_docstring(self):
        """
        Returns a docstring for this descriptor based on the corresponding
        docstring for :class:`django.db.models.QuerySet`.
        """
        docstring = self.get_base_docstring().strip()

        queryset_attr = getattr(QuerySet, self.name, None)
        queryset_doc = getattr(queryset_attr, '__doc__', '')
        if queryset_doc:
            if docstring:
                docstring += '  Documentation for *{name}*:\n'
            docstring += queryset_doc

        if inspect.ismethod(queryset_attr):
            docstring = '{}{}\n{}'.format(
                self.name,
                get_formatted_function_signature(queryset_attr),
                docstring
            )
        return docstring.strip().format(name=self.name)


class DelayedQuerySetMethod(DelayedQuerySetDescriptor):
    """
    A descriptor which acts like a method on a class.  When accessed
    as an attribute on *obj*, it returns its ``__call__`` method with
    *obj* being passed in as the first argument.
    """
    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        return partial(self.__call__, obj)


class PostApplyMethod(DelayedQuerySetMethod):
    """
    When this descriptor is called, it runs :meth:`DelayedQuerySet._apply`
    first, and then calls the corresponding method on that result of
    that operation.

    For example, with a :class:`DelayedUnionQuerySet`, when we call
    ``count()`` on it, then this descriptor will first take the union
    of its component querysets and then call ``.count()`` on the result.
    """
    def __call__(self, obj, *args, **kwargs):
        return getattr(obj._apply(), self.name)(*args, **kwargs)

    def get_base_docstring(self):
        return """
        Returns the output of ``{name}(...)`` after having applied the delayed
        operation.
        """


class PostApplyProperty(DelayedQuerySetDescriptor):
    """
    When this descriptor is called, it runs :meth:`DelayedQuerySet._apply`
    first, and then returns the corresponding property on the result of
    that operation.

    For example, with a :class:`DelayedUnionQuerySet`, when we access
    ``db`` on it, then this descriptor will first take the union
    of its component querysets and then return the ``db`` property on the
    result.
    """
    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        return getattr(obj._apply(), self.name)

    def get_base_docstring(self):
        return ""


class PassthroughMethod(DelayedQuerySetMethod):
    """
    When this descriptor is called, it calls the corresponding operation
    on all of the component querysets and then returns a clone of *obj*
    with those querysets.

    For example, when ``filter()`` is called on a :class:`DelayedUnionQuerySet`,
    it calls the corresponding filter method on all of the component
    querysets and then returns a :class:`DelayedUnionQuerySet` using
    those.
    """
    def __call__(self, obj, *args, **kwargs):
        return obj._clone([
            getattr(qs, self.name)(*args, **kwargs) for qs in obj._querysets
        ])

    def get_base_docstring(self):
        return """
        Returns the a new delayed queryset with ``{name}(...)`` having been called
        on each of the component querysets.:
        """


class FirstQuerySetPassthroughMethod(DelayedQuerySetMethod):
    """
    When this descriptor is called, returns a :class:`DelayedQuerySet`
    where the corresponding method has been applied to just the first
    component queryset.

    For example, when doing ``prefetch_related`` on a
    :class:`DelayedUnionQuerySet`, we only need to call ``prefetch_related``
    on the first queryset.  After the union is applied, then Django will
    use the prefetches from just that first queryset.
    """
    def __call__(self, obj, *args, **kwargs):
        querysets = (
            getattr(obj._querysets[0], self.name)(*args, **kwargs),
        ) + obj._querysets[1:]
        return obj._clone(querysets)

    def get_base_docstring(self):
        return """
        Returns the a new delayed queryset with ``{name}(...)`` having been
        called on the first component queryset, while the rest remain unchanged.
        """


class FirstQuerySetMethod(DelayedQuerySetMethod):
    """
    When this descriptor is called, returns the result when calling the
    corresponding method on the first queryset.
    """
    def __call__(self, obj, *args, **kwargs):
        return getattr(obj._querysets[0], self.name)(*args, **kwargs)

    def get_base_docstring(self):
        return """
        Returns the result of calling ``{name}(...)`` on the first component
        queryset.
        """


class NotImplementedMethod(DelayedQuerySetMethod):
    """
    A descriptor which raises a :class:`NotImplementedError` when called.
    """
    def __call__(self, obj, *args, **kwargs):
        raise NotImplementedError()

    def get_base_docstring(self):
        return """
        Raises :class:`NotImplementedError`.
        """


class DelayedQuerySetBase(abc.ABCMeta):
    """
    This is the metaclass for :`DelayedQuerySet`.  It's purpose is to make
    sure that the names are set for all :class:`DelayedQuerySetDescriptor`
    instances in the class.
    """
    def __new__(cls, name, bases, attrs):
        for key, value in attrs.items():
            if isinstance(value, DelayedQuerySetDescriptor):
                value.set_name(key)
        return super(DelayedQuerySetBase, cls).__new__(cls, name, bases, attrs)


class DelayedQuerySet(with_metaclass(DelayedQuerySetBase, object)):
    """
    A class used to work around some of the issues with Django's built-in
    support for set operations with querysets (such as ``UNION``).
    The primary issue is that after ```.union()`` call is made any subsequent
    filtering will silently fail. This class works around that issue by
    maintaing all of the individual querysets and not applying an operation
    like ``.union()`` until it's needed.

    For example, suppose we have ``qs = DelayedUnionQuerySet(qs0, qs1)```,
    then running ``qs = qs.filter(id=42)`` will be equivalent to doing
    ``qs = DelayedUnionQuerySet(qs0.filter(id=42), qs1.filter(id=42))``. Then,
    when we actually need to evaluate the queryset say by doing
    ``obj = qs.first()``, it will return ``qs0.union(qs1).first()`` behind
    the scenes.

    Subclasses need to implement the :meth:`_apply_operation`, which performs
    the operation such as ``.union()`` that is being delayed.
    """
    def __init__(self, *querysets, **kwargs):
        """
        :param tuple querysets: the component querysets
        :param dict kwargs: these are captured and made available for use
           in subclasses
        """
        if not all(isinstance(qs, QuerySet) for qs in querysets):
            # I think this should be able to work with DelayedQuerySets
            # as well, but the _apply method would need to call _apply
            # on any DelayedQuerySet in self._querysets.  Additionally,
            # only certain database engines support nested set operations.
            raise ValueError('can only pass in QuerySets for now')
        self._querysets = tuple(qs.order_by() for qs in querysets)
        self._kwargs = kwargs
        self._standard_ordering = True
        self._order_by = ()

    def _apply(self):
        """
        Returns a :class:`django.db.models.QuerySet` where
        :meth:`_apply_operation` has been applied and the global ordering
        has been set.

        :rtype: :class:`django.db.models.QuerySet`
        """
        qs = self._apply_operation().order_by(*self._order_by)
        qs.query.standard_ordering = self._standard_ordering
        return qs

    @abc.abstractmethod
    def _apply_operation(self):
        """
        Returns a proper :class:`django.db.models.QuerySet` from the
        component queryset.  This is the operation that
        :class:`DelayedQuerySet` is delaying.

        For example, :class:`DelayedUnionQuerySet`
        runs :meth:`django.db.models.QuerySet.union` with all of the
        component querysets.

        :rtype: :class:`django.db.models.QuerySet`
        """

    @property
    def model(self):
        """
        Returns the model class for the :class:`DelayedQuerySet`.
        """
        return self._querysets[0].model

    def _clone(self, querysets=None):
        """
        Returns a copy of this :class:`DelayedQuerySet` with the ordering
        preserved.

        :parama querysets: an optional iterable of querysets to use in the
           in the clone
        """
        if querysets is None:
            querysets = [qs._clone() for qs in self._querysets]
        clone = type(self)(*querysets, **self._kwargs)
        clone._order_by = self._order_by
        clone._standard_ordering = self._standard_ordering
        return clone

    __repr__ = PostApplyMethod()
    __len__ = PostApplyMethod()
    __iter__ = PostApplyMethod()
    __bool__ = PostApplyMethod()
    __nonzero__ = PostApplyMethod()
    __getitem__ = PostApplyMethod()

    __deepcopy__ = NotImplementedMethod()
    __getstate__ = NotImplementedMethod()
    __setstate__ = NotImplementedMethod()
    __and__ = NotImplementedMethod()
    __or__ = NotImplementedMethod()

    iterator = PostApplyMethod()
    count = PostApplyMethod()
    earliest = PostApplyMethod()
    latest = PostApplyMethod()
    first = PostApplyMethod()
    last = PostApplyMethod()
    delete = PostApplyMethod()
    exists = PostApplyMethod()
    none = PassthroughMethod()
    raw = PostApplyMethod()

    db = PostApplyProperty()

    all = PassthroughMethod()
    filter = PassthroughMethod()
    exclude = PassthroughMethod()
    values = PassthroughMethod()
    values_list = PassthroughMethod()
    annotate = PassthroughMethod()
    select_related = PassthroughMethod()
    defer = PassthroughMethod()
    only = PassthroughMethod()
    extra = PassthroughMethod()
    using = PassthroughMethod()
    complex_filter = PassthroughMethod()

    prefetch_related = FirstQuerySetPassthroughMethod()
    as_manager = FirstQuerySetPassthroughMethod()

    create = FirstQuerySetMethod()
    bulk_create = FirstQuerySetMethod()

    # These are left as not implemented at the moment.
    # We explicity put it here so that it is obvious to
    # the user of DelayedQuerysSet why things are not working.
    distinct = NotImplementedMethod()
    aggregate = NotImplementedMethod()
    union = NotImplementedMethod()
    intersection = NotImplementedMethod()
    difference = NotImplementedMethod()
    update = NotImplementedMethod()
    select_for_update = NotImplementedMethod()
    get_or_create = NotImplementedMethod()
    update_or_create = NotImplementedMethod()

    # These rely on sorting on an annotated field which is not available
    # once the union is applied
    dates = NotImplementedMethod()
    datetimes = NotImplementedMethod()

    def get(self, *args, **kwargs):
        """
        Performs the query and returns a single object matching the given
        keyword arguments.

        .. note::

           We cannot use :class:`PostApplyMethod` for this since that does
           additional filtering which does not work with querysets that have
           been "unioned" for example.
        """
        clone = self.filter(*args, **kwargs)
        values = list(clone)
        num = len(values)
        if num == 1:
            return values[0]

        if not num:
            raise self.model.DoesNotExist(
                "%s matching query does not exist." %
                self.model._meta.object_name
            )
        raise self.model.MultipleObjectsReturned(
            "get() returned more than one %s -- it returned %s!" %
            (self.model._meta.object_name, num)
        )

    def order_by(self, *field_names):
        """
        Returns a new :class:`DelayedQuerySet`` instance with the ordering
        changed.

        .. note::

           We need to have a custom implementation for this because we
           want to change the ordering of the final queryset, not just
           the ordering within each component queryset.
        """
        qs = self._clone()
        qs._order_by = field_names
        return qs

    @property
    def ordered(self):
        """
        Returns True if the :class:`DelayedQuerySet` is ordered -- i.e.
        has an order_by() clause.

        :rtype: bool
        """
        return bool(self._order_by)

    def reverse(self):
        """
        Reverses the ordering of the :class:`DelayedQuerySet`.

        .. note::

           We need to have a custom implementation for this because we
           want to reverse the ordering of the final queryset, not just
           the ordering within each component queryset.
        """
        qs = self._clone()
        qs._standard_ordering = not qs._standard_ordering
        return qs

    def in_bulk(self, id_list=None):
        """
        Returns a dictionary mapping each of the given IDs to the object with
        that ID. If `id_list` isn't provided, the entire
        :class:`DelayedQuerySet` is evaluated.
        """
        if id_list is not None:
            if not id_list:
                return {}
            qs = self.filter(pk__in=id_list).order_by()
        else:
            qs = self._clone()
        return {obj._get_pk_val(): obj for obj in qs}



import inspect
import re

from future.utils import PY2


def get_formatted_function_signature(func):
    if PY2:
        signature = inspect.formatargspec(*inspect.getargspec(func))
    else:
        signature = inspect.formatargspec(*inspect.getfullargspec(func))
    return re.sub('self(, )?', '', signature).replace('()', '( )')



from django.test import TestCase

from django_delayed_union import DelayedUnionQuerySet
from django_delayed_union.base import DelayedQuerySetBase
from django_delayed_union.base import DelayedQuerySetDescriptor


class DelayedQuerySetBaseTests(TestCase):
    def setUp(self):
        super(DelayedQuerySetBaseTests, self).setUp()
        self.cls = DelayedUnionQuerySet

    def test_has_metaclass(self):
        self.assertTrue(
            issubclass(type(self.cls), DelayedQuerySetBase)
        )

    def test_descriptors_have_names_set(self):
        for name in dir(self.cls):
            descriptor = getattr(self.cls, name)
            if not isinstance(descriptor, DelayedQuerySetDescriptor):
                continue
            self.assertEqual(descriptor.name, name)



import os

SECRET_KEY = 'roverdotcom'
INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
]
ROOT_URLCONF = []

TEST_DATABASE = os.environ.get('TEST_DATABASE', 'sqlite')

if TEST_DATABASE == 'mysql':
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.mysql',
            'NAME': 'testdb',
            'USER': 'travis',
            'PASSWORD': '',
            'HOST': '127.0.0.1',
            'PORT': 3306,
        }
    }
if TEST_DATABASE == 'postgresql':
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql_psycopg2',
            'NAME': 'testdb',
            'USER': 'postgres',
            'PASSWORD': '',
            'HOST': '127.0.0.1',
            'PORT': 5432,
        }
    }
else:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3'
        }
    }



from .base import DelayedQuerySet


class DelayedUnionQuerySet(DelayedQuerySet):
    def __init__(self, *querysets, **kwargs):
        kwargs.setdefault('all', False)
        unexpected_kwarg = next((k for k in kwargs.keys() if k != 'all'), None)
        if unexpected_kwarg:
            raise TypeError(
                "received an unexpected keyword argument '{}'".format(
                    unexpected_kwarg
                )
            )
        return super(DelayedUnionQuerySet, self).__init__(*querysets, **kwargs)

    def _apply_operation(self):
        """
        Returs the union of all of the component querysets.
        """
        return self._querysets[0].union(*self._querysets[1:], **self._kwargs)

    def distinct(self):
        """
        Returns a new :class:`DelayedUnionQuerySet` instance that will
        select only distinct results.
        """
        clone = self._clone()
        clone._kwargs['all'] = False
        return clone

    def update(self, **kwargs):
        """
        Updates all elements in the component querysets, setting all the given
        fields to the appropriate values.  Returns the total number of
        (not-necessarily distinct) rows updated.
        """
        total_count = 0
        for queryset in self._querysets:
            total_count += queryset.update(**kwargs)
        return total_count



from django.contrib.auth.models import Permission
from django.contrib.auth.models import User
from django.test import TestCase

from django_delayed_union import DelayedDifferenceQuerySet

from .factories import UserFactory
from .mixins import DelayedQuerySetMetaTestsMixin
from .mixins import DelayedQuerySetTestsMixin


class DelayedDifferenceQuerySetMetaTests(
        DelayedQuerySetMetaTestsMixin,
        TestCase):

    def get_class(self):
        return DelayedDifferenceQuerySet


class DelayedDifferenceQuerySetTestsMixin(DelayedQuerySetTestsMixin):
    pass


class DelayedDifferenceQuerySetTests(
        DelayedDifferenceQuerySetTestsMixin,
        TestCase):

    @classmethod
    def setUpTestData(cls):
        super(DelayedDifferenceQuerySetTests, cls).setUpTestData()
        cls.excluded_user = UserFactory.create()

    def get_queryset(self):
        return DelayedDifferenceQuerySet(
            User.objects.all(),
            User.objects.filter(id=self.excluded_user.id)
        )

    def test_select_related(self):
        base_qs = Permission.objects.all()
        qs = DelayedDifferenceQuerySet(base_qs, base_qs.none())

        self.assertTrue(qs.exists())
        for permission in qs.select_related('content_type'):
            with self.assertNumQueries(0):
                self.assertIsNotNone(permission.content_type.id)



from .difference import DelayedDifferenceQuerySet
from .intersection import DelayedIntersectionQuerySet
from .union import DelayedUnionQuerySet

__version__ = '0.1.0'

__all__ = [
    '__version__',
    'DelayedDifferenceQuerySet',
    'DelayedIntersectionQuerySet',
    'DelayedUnionQuerySet',
]
